#
# HocusPocus Alarm Clock Plugin
#
# Copyright (C)  2010 and onwards Robert Muth <robert at muth dot org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# python imports
import datetime
import logging
import os
import threading
import time
# local imports
import utils


def CheckedOrEmpty(x):
    if x:
        return "checked"
    else:
        return ""



class Alarm:
    def __init__(self):
        self.h = 0
        self.m = 0
        self.days = [False] * 7
        return

    def MakeBlob(self, no, refresh):
         obj = utils.Datablob()
         obj.no = no
         obj.h = self.h
         obj.m = self.m
         obj.action = "alarm"
         obj.day0 = CheckedOrEmpty(self.days[0])
         obj.day1 = CheckedOrEmpty(self.days[1])
         obj.day2 = CheckedOrEmpty(self.days[2])
         obj.day3 = CheckedOrEmpty(self.days[3])
         obj.day4 = CheckedOrEmpty(self.days[4])
         obj.day5 = CheckedOrEmpty(self.days[5])
         obj.day6 = CheckedOrEmpty(self.days[6])
         return obj

    def ComputeNextAlarm(self, dt):
        oneday = datetime.timedelta(1)
        al = datetime.datetime(dt.year, dt.month, dt.day, self.h, self.m)
        if al <= dt:
            al += oneday
        assert al > dt
        for i in range(7):
            weekday = al.weekday()
            if self.days[weekday]:
                return [al]
            al += oneday
        return []

    def Update(self, args):
       self.days[0] =  "day0" in args
       self.days[1] =  "day1" in args
       self.days[2] =  "day2" in args
       self.days[3] =  "day3" in args
       self.days[4] =  "day4" in args
       self.days[5] =  "day5" in args
       self.days[6] =  "day6" in args
       self.h = int(args["h"])
       self.m = int(args["m"])

class AlarmManager:
    def AlarmCheckerThread(self):
        logging.info("alarm thread started")
        while self._keep_going:

            time.sleep(self._interval)
            if not self._next_alarm:
                continue
            now = datetime.datetime.now()
            if now < self._next_alarm:
                continue
            self.StartAlarm()


    def ComputeNextAlarm(self):
        now = datetime.datetime.now()
        next_alarms = []
        for a in self._alarms:
            next_alarms += a.ComputeNextAlarm(now)
        if not next_alarms:
            return None
        else:
            return min(next_alarms)

    def StartAlarm(self):
        self._alarm_active = True
        self._next_alarm = None
        logging.info("start alarm")
        utils.RunCommand(self._start_cmd)
        self._alarm_active = False

    def StopAlarm(self):
        logging.info("stop alarm")
        utils.RunCommand(self._stop_cmd)
        self._next_alarm = self.ComputeNextAlarm()


    def __init__(self, no, interval, start_cmd, stop_cmd):
        self._alarm_active = False
        self._alarms = [Alarm() for _ in range(no)]
        self._start_cmd = start_cmd
        self._stop_cmd = stop_cmd
        self._keep_going = True
        self._next_alarm = None
        self._interval = interval
        self._thread = threading.Thread(target=self.AlarmCheckerThread)
        self._thread.setDaemon(True)
        self._thread.start()


    def MakeDataBlobs(self, refresh):
        return [a.MakeBlob(no, refresh)
                           for no,a  in enumerate(self._alarms)]


    def Update(self, args):
        assert "no" in args
        no = int(args["no"])
        assert no < len(self._alarms)
        self._alarms[no].Update(args)
        self._next_alarm = self.ComputeNextAlarm()

class Handler:

    def __init__(self, util, config):
        self._util = util
        self._config = self._util.LoadConfig("alarm.conf")
        self._alarm_manager = AlarmManager(self._config.num_alarms,
                                           self._config.sleep_interval,
                                           self._config.start_cmd,
                                           self._config.stop_cmd
                                           )
        return



    def Handler(self, req, path, args):
        cmd = args.get("cmd", "")
        refresh = int(utils.EscapeContents(args.get("refresh",
                                                    str(self._config.refresh))))
        logging.info("Handler: alarm %s", cmd)
        if not cmd:
            pass
        elif cmd == "update":
            self._alarm_manager.Update(args)
        elif cmd == "test":
            self._alarm_manager.StartAlarm()
        elif cmd == "stop":
            self._alarm_manager.StopAlarm()
        else:
            logging.error("unknown command")
        base_url = "alarm?refresh=%d" % refresh
        data= self._util.GetDefaultDict()
        now = datetime.datetime.now()

        data['now'] = now.strftime(self._config.time_format)
        next_alarm = self._alarm_manager._next_alarm
        if next_alarm:
            data['next_alarm'] = next_alarm.strftime(self._config.time_format)
            delta = next_alarm - now
            data['next_alarm'] += " [%s]" % str(delta)
        else:
            data['next_alarm'] = "No alarm set"

        data['alarm_active'] = self._alarm_manager._alarm_active
        data['hours'] =  [str(i) for i in range(0, 24)]
        data['minutes'] =  [str(i) for i in range(0, 60, 5)]
        data['alarms'] = self._alarm_manager.MakeDataBlobs(refresh)

        data['refresh-time'] = str(refresh)
        data['refresh-url'] = base_url
        data['refreshs'] = utils.MakeBlobRefreshList("alarm?refresh=%d",
                                                     self._config.refreshs,
                                                     refresh)
        data['plugin-name'] = "Alarm"

        self._util.CompleteRequest(req,
                                   data,
                                   self._config.template)
        return

