#
# HocusPocus Audacious Plugin
#
# Copyright (C)  2008 and onwards Robert Muth <robert at muth dot org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# python imports
import logging
import os
# local imports
import utils

for t in ["/usr/bin/audtool2", "/usr/bin/audtool"]:
    if os.path.exists(t):
        AUDTOOL = t
        break

for t in ["/usr/bin/audacious2", "/usr/bin/audacious"]:
    if os.path.exists(t):
        AUDACIOUS = t
        break

COMMAND = {
    'play':  [AUDTOOL, '--playback-play'],
    'stop':  [AUDTOOL, '--playback-stop']    ,
    'pause': [AUDTOOL, '--playback-playpause'],
    'next':  [AUDTOOL, '--playlist-advance'],
    'prev':  [AUDTOOL, '--playlist-reverse'],
    'fwd+':  [AUDTOOL, '--playback-seek-relative', '1'],
    'ffwd':  [AUDTOOL, '--playback-seek-relative', '10'],
    'fffwd': [AUDTOOL, '--playback-seek-relative', '100'],
    'rev':   [AUDTOOL, '--playback-seek-relative', '-1'],
    'rrev':  [AUDTOOL, '--playback-seek-relative', '-10'],
    'rrrev': [AUDTOOL, '--playback-seek-relative', '-100'],
    'get_volume':   [AUDTOOL, '--get-volume'],
    'set_volume':   [AUDTOOL, '--set-volume', '%(vol)s'],
    'curr_song':   [AUDTOOL, '--current-song'],
    'playlist_pos':   [AUDTOOL, '--playlist-position'],
    'get_playlist':   [AUDTOOL, '--playlist-display'],
    'curr_dur': [AUDTOOL, '--current-song-length-seconds'],
    'curr_pos':    [AUDTOOL, '--current-song-output-length-seconds'],
    'shuffle_status':    [AUDTOOL, '--playlist-shuffle-status'],
    'shuffle_toggle':    [AUDTOOL, '--playlist-shuffle-toggle'],
    'kill': ["killall", "-9", AUDACIOUS],
    }


def RunAudTool(name, args={}):
    comm = COMMAND.get(name, None)
    if not comm:
        logging.error("bad command %s", name)
        return -1, ""

    return utils.RunCommand([c % args for c in comm])


def GetPlaylist():
    result, pos = RunAudTool("playlist_pos")
    if result != 0:
        return None, None
    pos = int(pos)
    result, playlist_raw = RunAudTool("get_playlist")
    if result != 0:
        return None, None
    playlist_raw = playlist_raw.split("\n")

    def clean(line):
        start = line.find("|")
        if start < 0:
            return ""
        end = line.rfind("|")
        return line[start + 1: end].strip()

    playlist = [clean(line) for line in playlist_raw]
    playlist = [p for p in playlist if p]

    return playlist, pos - 1


def MakeSongBlob(title, pos=0, dur=0):
    song = utils.Datablob()
    song.title = title
    song.dur = dur
    song.pos = pos
    if int(dur):
        n = 20 * int(pos) / int(dur)
        song.progress = (utils.MakeHtmlBulletsFull(n) +
                         utils.MakeHtmlBulletsEmpty(20 - n))

    else:
        song.progress = ""
    return song


def GetShuffleStatus():
    result, mode = RunAudTool("shuffle_status")
    if mode.startswith("on"):
        return "on"
    else:
        return "off"


def GetVolume():
    result, output = RunAudTool("get_volume")
    try:
        return int(output)
    except:
        return 0

class Handler:

    def __init__(self, util, config):
        self._util = util
        self._config = self._util.LoadConfig("audacious.conf")
        return

    def MakeSongBlobs(self):
        songs = []
        playlist, pos = GetPlaylist()
        if not playlist:
            songs.append(MakeSongBlob("No Song Playing", 0, 0))
        else:
            playlist = playlist[pos: pos + self._config.num_history]
            result, dur = RunAudTool("curr_dur")
            result, pos = RunAudTool("curr_pos")
            songs.append(MakeSongBlob(playlist[0], pos, dur))
            for s in playlist[1:]:
                songs.append(MakeSongBlob(s))
        return songs

    def MakeVolumeBlobs(self, base_url):
        current = GetVolume()
        vols = []
        steps = 100 / self._config.num_vol_levels
        for i in range(0, 101, steps):
            v = utils.Datablob()
            v.url = base_url + "&amp;cmd=set_volume&amp;vol=%d" % int(i)
            if i <= (current + steps - 1) < (i + steps):
                v.val = utils.MakeHtmlBulletsFull(1)
            else:
                v.val = utils.MakeHtmlBulletsEmpty(1)
            vols.append(v)
        return vols

    def Handler(self, req, path, args):
        cmd = args.get("cmd", "")
        refresh = int(utils.EscapeContents(args.get("refresh",
                                                    str(self._config.refresh))))
        logging.info("Handler: audacious %s", cmd)
        if cmd:
            RunAudTool(cmd, args)
        base_url = "audacious?refresh=%d" % refresh
        data= self._util.GetDefaultDict()
        data['songs'] =  self.MakeSongBlobs()
        data['vol_levels'] =  self.MakeVolumeBlobs(base_url)
        data['shuffle'] = GetShuffleStatus()
        data['refreshs'] = utils.MakeBlobRefreshList("audacious?refresh=%d",
                                                     self._config.refreshs,
                                                     refresh)
        data['refresh-time'] = str(refresh)
        data['refresh-url'] = base_url

        data['plugin-name'] = "Audacious"

        self._util.CompleteRequest(req,
                                   data,
                                   self._config.template)
        return

