#!/usr/bin/env python
#
# HocusPocus
#
# Copyright (C) Robert Muth <robert at muth dot org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# python import
import BaseHTTPServer
import SocketServer
import getopt
import glob
import logging
import mimetypes
import os
import sys
import time

# local imports
import authentication
import ezt
import utils

######################################################################
##
######################################################################

class TemplateCache(object):

    def __init__(self, dir, encodings):
        self._dir = dir
        self._encodings = encodings
        self._cache = {}

    def GetTemplate(self, name):
        no_caching = 1 # for debugging
        if no_caching or name not in self._cache:
            full_path = os.path.join(self._dir, name)
            self._cache[name] = ezt.Template(full_path, self._encodings)

        return self._cache[name]


######################################################################
##
######################################################################
class Util(object):
    """
    misc class for random stuff

    TODO: refactor this
    """
    def __init__(self, config):
        # cool
        """
        cool
        """
        self._stats_count = {}
        self._user_log = []
        self._config = config
        self._template_cache = TemplateCache(config.templates_path,
                                             config.templates_encodings)
        return

    def IncCounter(self, family, name, value=1):
        key = (family, name)
        self._stats_count[key] = value + self._stats_count.get(key, 0)
        return

    def LogUser(self, ip, timestamp, message):
        if len(self._user_log) > 50:
            self._user_log.pop(0)
        self._user_log.append((ip, timestamp, message))
        self.IncCounter("ip-access", ip)
        return

    def LoadConfig(self, fname):
        fname = os.path.join(self._config.configs_path, fname)
        mod, name = utils.LoadPythonModule(fname)
        return mod

    def GetDefaultFooter(self):
        return self._config.footer_template

    def GetDefaultHeader(self):
        return self._config.header_template

    def GetDefaultDict(self):
        def make_obj(url, name):
            blob = utils.Datablob()
            blob.url = url
            blob.name = name
            return blob

        def maybe_add_padding(s):
            if len(s) <= 3:
                return "&nbsp;" + s + "&nbsp;"
            else:
                return s

        data = {
            'plugin-name': "change-me-name",
            'refresh-time': "",
            'refresh-url': "",
            'prefix': self._config.name,
            'tools': [make_obj(url, maybe_add_padding(name))
                       for url, name in self._config.tools_links],
            }
        return data

    def CompleteRequest(self, req,
                        data,
                        template=None,
                        code=200,
                        type='text/html',
                        location=None,
                        length=None,
                        mtime=None):

        req.send_response(code)
        req.send_header("Content-Type", type)
        if location:
            req.send_header("Location", location)
        if length:
            req.send_header("Content-Length", length)
        if mtime:
            req.send_header('Last-Modified',
                            time.strftime("%a, %d %b %Y %T GMT"))
        req.end_headers()
        if template:
            head = self._template_cache.GetTemplate("hocuspocus.head.ezt")
            body = self._template_cache.GetTemplate(template)
            foot = self._template_cache.GetTemplate("hocuspocus.foot.ezt")
            head.generate(req.wfile, data)
            body.generate(req.wfile, data)
            foot.generate(req.wfile, data)
        else:
            req.wfile.write("".join(data))
        return

    def ErrorPage(self, req, message):
        data = self.GetDefaultDict()
        data['message'] = message
        self.CompleteRequest(req, data,  "error.ezt")
        return

######################################################################
##
######################################################################

class ResourceHandler(object):

    def __init__(self, util, config):
        self._util = util
        self._config = config
        return

    def Handler(self, req, path, args):
        logging.info("Handler: resource %s", str(path))
        if len(path) <=1 :
            data = ["missing arg<hr/>"]
            self._util.CompleteRequest(req, data, None, 404)
            return
        filename = os.path.join(self._config.resources_path, path[1])
        type = mimetypes.guess_type(filename)[0]
        if not type:
            data = ["unknown mimetype for %s<hr/>" % filename]
            self._util.CompleteRequest(req, data, None, 404)
            return

        f = open(filename, 'rb')
        st = os.fstat(f.fileno())
        clen = st.st_size
        mtime = st.st_mtime
        data = [f.read()]
        f.close()
        self._util.CompleteRequest(req, data, None,
                                 type=type, length=clen, mtime=mtime);

        return

######################################################################
##
######################################################################
class TestHandler(object):
    def __init__(self, util):
        self._util = util
        return

    def Handler(self, req, path, args):
        logging.info("test handler invoked")

        self._util.CompleteRequest(req,
                                   "OK\n",
                                   None,
                                   type="text/plain")
        return
######################################################################
##
######################################################################
class IdHandler(object):
    """Docu string for class"""

    def __init__(self, util, config):
        self._util = util
        self._name = config.name
        return

    def Handler(self, req, path, args):
        logging.info("id handler invoked")

        self._util.CompleteRequest(req,
                                   self._name + "\n",
                                   None,
                                   type="text/plain")
        return
######################################################################
##
######################################################################
class ReloadHandler(object):
    """Docu string for class"""

    def __init__(self, util, config):
        self._util = util
        self._name = config.name
        return

    def Handler(self, req, path, args):
        logging.info("reload handler invoked")
        req.server.ReloadHandlers()
        self._util.CompleteRequest(req,
                                   "Reload Complete\n",
                                   None,
                                   type="text/plain")
        return

######################################################################
##
######################################################################
class DefaultHandler(object):

    def __init__(self, util, config):
        self._util = util
        self._config = config
        return

    def Handler(self, req, path, args):
        def make_obj(url, name):
            blob = utils.Datablob()
            blob.url = url
            blob.name = name
            return blob

        logging.info("default handler invoked %s", str(path))
        data = self._util.GetDefaultDict()
        data['plugins'] = [make_obj(*p) for p in self._config.menu_links]
        data["plugin-name"] = "Main Menu"
        self._util.CompleteRequest(req,
                                   data,
                                   self._config.default_template)
        return
######################################################################
##
######################################################################
class AboutHandler(object):

    def __init__(self, util, config):
        self._util = util
        self._config = config
        return

    def Handler(self, req, path, args):
        logging.info("about handler invoked %s", str(path))
        data = self._util.GetDefaultDict()
        data["plugin-name"] = "About"
        self._util.CompleteRequest(req,
                                   data,
                                   self._config.about_template)
        return


######################################################################
##
######################################################################
class RequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def IncCounter(self, a, b):
        self.server._util.IncCounter(a, b)

    def SendAuthenticationRequest(self, data):
        realm='hocuspocus'
        self.send_response(401)
        self.send_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
        self.send_header('Content-Type', 'text/html;');
        self.send_header('Connection', 'close');
        self.end_headers()
        self.wfile.write(data)
        return

    def do_GET(self):
        authenticator = authentication.Authenticator(self.server._config.auth_machines,
                                                         self.server._config.auth_users)
        t = authenticator.Authenticate(self)
        if not t:
            self.IncCounter("authenticate", "failure")
            if self.server._config.auth_users:
                logging.info("authentication request")
                self.IncCounter("authentication", "request")
                self.SendAuthenticationRequest("Authentication Required")
            return
        self.IncCounter("authenticate", t[0])
        self.IncCounter("authenticate", t[0] + "-" + t[1])

        if t[0] == "user":
            self.IncCounter("user", t[1])
        else:
            self.IncCounter("generic", t[1])

        path, args = self.translate_path()
        if path and path[0] == "favicon.ico":
            path = ["resources", "favicon.ico"]

        if not path or path[0] not in self.server._handlers:
            handler = "default"
        else:
            handler = path[0]
        #print "PATH: ", path
        #print "ARGS: ", args
        self.IncCounter("handler", handler)
        ip, port = self.client_address
        self.server._util.LogUser(ip, time.time(), path)
        self.server._handlers[handler](self, path, args)
        return

    def translate_path(self):
        args = {}
        path = self.path
        if "?" in path:
            path, params = path.split("?", 2)
            params = params.split("&")
            for p in params:
                if "=" in p:
                    k, v= p.split("=", 2)
                    args[k] = v
                else:
                  args[p] = None
        path = path.split("/")
        path = [x for x in path if x]
        return path, args

    def log_message(self, format, *args):
        logging.info("Served Request " + format % args)

######################################################################
##
######################################################################
def AddStandardHandlers(handlers, util, config):
    default_handler = DefaultHandler(util, config)
    handlers["default"] = default_handler.Handler
    resource_handler = ResourceHandler(util, config)
    handlers["resources"] = resource_handler.Handler
    test_handler = TestHandler(util)
    handlers["test"] = test_handler.Handler
    id_handler = IdHandler(util, config)
    handlers["id"] = id_handler.Handler
    reload_handler = ReloadHandler(util, config)
    handlers["reload"] = reload_handler.Handler
    about_handler = AboutHandler(util, config)
    handlers["about"] = about_handler.Handler
    return

def AddPluginHandlers(handlers, util, config):
    dir = config.plugins_path
    logging.info("loading plugins from %s", dir)
    desc =  ('.py', 'U', 1)
    for fname in glob.glob(os.path.join(dir,"*.py")):
        mod, name = utils.LoadPythonModule(fname)
        handler = mod.Handler(util, None)
        handlers[name] = handler.Handler
    return

class Server(SocketServer.ThreadingMixIn, BaseHTTPServer.HTTPServer):

    def ReloadHandlers(self):
        utils.UnloadAllPythonModules()
        self._handlers = {}
        AddStandardHandlers(self._handlers, self._util, self._config)
        AddPluginHandlers(self._handlers, self._util, self._config)

    def __init__(self, util, config):
        self._util = util
        self._config = config
        self._handler = {}
        self.ReloadHandlers()
        port = config.server_port

        logging.info("starting server on port %d", port)
        logging.info("*" * 70)
        logging.info("Hocuspocus ready, point your browser at   "
                     "http://localhost:%d", port)
        logging.info("*" * 70)
        SocketServer.TCPServer.__init__(self, ("", port), RequestHandler)

######################################################################
##
######################################################################
def ParseCommandline(argv):
    defaults = utils.Datablob()
    defaults.debug_level = logging.INFO
    defaults.config_file = "hocuspocus.conf"
    defaults.log_file = None

    try:
        opts, args = getopt.getopt(argv[1:], "f:d:l:")
    except getopt.GetoptError, err:
        return  str(err), defaults

    for o, a in opts:
        if o == "-d":
            defaults.defaults = int(a)
        elif o == "-f":
            defaults.config_file = a
        elif o == "-l":
            defaults.log_file = a
        else:
            assert False, "unhandled option"
    return "", defaults



def main(argv):
    global GlobalHandlers
    mess, defaults = ParseCommandline(argv)
    if mess:
        print mess
        return -1

    logformat = ("%(levelname)s:"
                 "%(module)s:"
                 "%(lineno)d:"
                 #             "%(threadName)s:"
                 " %(message)s")

    if defaults.log_file:
        logging.basicConfig(level=logging.INFO,
                            format=logformat,
                            filemode="w+",
                            filename=defaults.log_file)
    else:
        logging.basicConfig(level=logging.INFO,
                            format=logformat)
    logging.info("started log")

    config, dummy = utils.LoadPythonModule(defaults.config_file)

    util = Util(config)
    util.log_file = defaults.log_file

    server = Server(util, config)
    try:
        while 1:
            server.handle_request()
    except KeyboardInterrupt:
        logging.warning("\nCaught ctr-c, taking down the server")
        logging.warning("Please wait while the remaining streams finish..")
    logging.warning("exiting server")
    logging.shutdown()
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))
